#include "hash.hh"
#include "expr.hh"
#include "stmt.hh"

typedef autoptr<struct TypeInfo> TypeP;
typedef autoptr<struct UserType> UtypeP;
typedef autoptr<struct Scope> ScopeP;
typedef autoptr<struct SubRoutine> SubFuncP;

struct TypeInfo: public ptrable
{
    enum
    {
        Integer, Long, Single, Double, String,
        UserType
    } Type;
    
    bool dynamic;            //if resizable array
    unsigned fixlength;      //if non-dynamic string.
    UtypeP members;          //if UserType
    
    // If array:
    std::vector<std::pair<unsigned, unsigned> > dimensions;
    // ^first  = beginning index
    // ^second = dimension (size)
    //   example: 1 TO 5   is (1,5)
    //            0 TO 4   is (0,5)
    
    TypeInfo()
    : Type(Integer),dynamic(false),fixlength(0),members(),dimensions()
    { }
    
    bool is_string() const { return Type == String; }
    bool is_num() const { return Type != String && Type != UserType; }
    bool is_array() const { return !dimensions.empty(); }
    
    bool operator==(const TypeInfo& b) const;
    bool operator!=(const TypeInfo& b) const { return !operator==(b); }
};

struct UserType: public std::vector<std::pair<std::string, TypeP> >, public ptrable
{
    hash_set<std::string> Names;
    
    bool DeclareMember(const std::string& name, TypeP p);
    
    bool operator==(const UserType& b)const;
};

struct Constant
{
    QBtoken Type; //integer,single,double,long or string.
    exprp value;
    

    Constant(): Type((QBtoken)0),value() {}
};

struct Subroutine // SUB, FUNCTION, DEF
{
    stmtp   code;
    ScopeP scope;
};

struct TypeFlagSet
{
    bool is_subroutine:1;
    bool is_scalar    :1;
    bool is_const     :1;
    bool is_array     :1;
    
    bool is_global_scalar : 1; // imported using SHARED
    bool is_global_array  : 1; // imported using SHARED
    
    // In case of subroutines, they can be DECLARED and later defined.
    bool has_body     :1;
    
    // Types and labels not included here; they are in different namespace.
    
    TypeFlagSet() {}
};

typedef autoptr<struct VarContextDef> ctxp;
struct VarContextDef: public ptrable
{
    TypeP type;
    bool is_static; // STATIC to the given function
    bool is_shared; // SHARED to the subfunctions (towards children)
    
public:
    VarContextDef(): type(),is_static(false),is_shared(false) {}
};

/* SUB, FUNCTION, DEF FN */
struct SubRoutine: public ptrable
{
    struct ParamSpec
    {
        std::string name;
        TypeP       type;
        enum { ByRef, ByVal, Seg } passconvention;
        
        ParamSpec() : name(), type(), passconvention(ByRef) { }
        
        void SetByval() { passconvention=ByVal; }
        void SetSeg() { passconvention=Seg; }
    };
    
    TypeP returntype; // 0=SUB, otherwise FUNCTION/DEF
    std::vector<ParamSpec> params;
    bool is_static;
    stmtp code;
    
    int callconvention;     // 0=default, 1=cdecl
    std::string linkername; // ALIAS

public:
    SubRoutine(): returntype(), params(), is_static(false),
                  code(), callconvention(0), linkername() {}
};

typedef hash_map<std::string, ctxp> ScopeVarListType;

struct Scope: public ptrable
{
    /*****
      Method of lookup:
        Check constants. Ignore suffix.
          Match? Fail if had wrong suffix. Otherwise, accept.
      
        Check UTYPE.
          Match? Fail if had a suffix. Otherwise, accept.
        Had suffix? Check by suffix.
        Otherwise, check by DEFtype.
    *******/
    
    hash_map<std::string, UtypeP> types;       // Types have no suffix.
    hash_map<std::string, Constant> constants; // Constants have no suffix.
    
    // SUBs, FUNCTIONs, DEFs
    hash_map<std::string, SubFuncP> subroutines;
    
    struct ScopeDef
    {
        ScopeVarListType scalars;
        ScopeVarListType arrays;
    } Utype, Int, Single, Double, Long, String;
    
    hash_map<std::string, TypeFlagSet> usage;
    
    // Labels...
    hash_set<std::string> labels;
};
